;; Programs for viewing galaxies.  These are intended for use with
;; either gplotc.lisp or gplott.lisp, and refer to global variables
;; defined therein to control viewpoint, direction, and luminosity
;; calculations.

;; In order to simplify changing the viewing situation, the program
;; introduces a "view center", viewc, a vector in space coordinates
;; at which the view is directed, so that
;;
;;     pc - viewc is parallel to pn, and in the same sense
;;
;; Rotations are of two classes, those that preserve viewc and move pc,
;; and those which preserve pc and move viewc.  Tranlations always move
;; pc, and may or may not move viewc.

;; The magnitude of motions is determined by deltax and deltath, for
;; translations and rotations respectively.  These may be specifed
;; with any command which uses them.  Otherwise, the most recent value
;; is used.

;; The commands: (parameters are optional)
;;
;;   p <degrees>  - rotate pc up
;;   n <degrees>  - rotate pc down
;;   l <degrees>  - rotate pc left
;;   r <degrees>  - rotate pc right
;;   f <distance> - move pc forward, viewc fixed
;;   b <distance> - move pc backward, viewc fixed
;;
;;   P <degrees>  - rotate viewc up
;;   N <degrees>  - rotate viewc down
;;   L <degrees>  - rotate viewc left
;;   R <degrees>  - rotate viewc right
;;   F <distance> - move pc forward, viewc moving
;;   B <distance> - move pc backward, viewc moving
;;
;;   u <vector>   - rotate screen so <vector> is up
;;   d <length>   - adjust standard luminosity distance to <length>
;;   z,Z          - move viewc and pc so viewc = #(0 0 0)
;;   s,S          - standard pos'n (determined when used)
;;   w,W          - where am I? (issues report)
;;
;;   e <expr> - evaluate <expr> and print result (also E)
;;   q,Q      - quit

(in-package '*lisp)

(defvar viewc #(0.0 0.0 0.0))  ;; "View center"
(defvar viewcd 0.0)            ;; Distance viewc - pc
(defvar deltax 10.0)           ;; distance increment - space units
(defvar deltath 0.1745329252)  ;; angle increment - radians (10 deg.)
(defvar vipc #(1.0 0.0 0.0))   ;; unit vector viewc -> pc
(defvar viup #(0.0 0.0 1.0))   ;; unit vector -vj (up direction)
(defvar virt #(0.0 1.0 0.0))   ;; unit vector vi (right)

;; Standard values

(defvar std-viewc viewc)
(defvar std-pc pc)
(defvar std-viup viup)

;; Calculate current view situation

(defun viewstate ()
  (setq vipc (v- pc viewc))
  (setq viewcd (vabs vipc))
  (setq vipc (v/-constant vipc viewcd))
  (setq viup (vscale-to-unit-vector (v- vj)))
  (setq virt (vscale-to-unit-vector vi))
)

;; Functions which calculate new views leave results in pc, pn,
;; and viup.  pn and viup are calculated only for direction.
;; They assume the view situation has already been calculated.

;; Calculate new view for rotating (revolving) pc up or down
;; th > 0 is up

(defun pupdn (th)
 (declare (type single-float th))
  (let ((s (sin th)) (c (cos th))) (declare (type single-float s c))
    (setq pn (v+ (v*-constant vipc c) (v*-constant viup s)))
    (setq pc (v+ viewc (v*-constant pn viewcd)))
    (setq viup (v- (v*-constant viup c) (v*-constant vipc s)))
  )
)

;; Calculate for rotating pc right (th > 0) or left

(defun prtlf (th)
 (declare (type single-float th))
  (let ((s (sin th)) (c (cos th))) (declare (type single-float s c))
    (setq pn (v+ (v*-constant vipc c) (v*-constant virt s)))
    (setq pc (v+ viewc (v*-constant pn viewcd)))
  )
)

;; Calculate for moving pc forward (dx > 0) or back, viewc fixed

(defun pfb (dx)
 (declare (type single-float dx))
  (setq pc (v+ pc (v*-constant vipc (- dx))))
)

;; Swap pc and viewc for motions of viewc

(defun viewswap ()
 (let ((temp pc))
  (setq pc viewc) (setq viewc temp) (setq pn (v- pn))
  (setq vipc (v- vipc)) (setq virt (v- virt))
 )
)

;; Calculate for moving viewc up (th > 0) or down

(defun Vupdn (th)
 (declare (type single-float th))
   (viewswap) (pupdn th) (viewswap)
)

;; Calculate for moving viewc right (th > 0) or left

(defun Vrtlf (th)
 (declare (type single-float th))
   (viewswap) (prtlf (- th)) (viewswap)
)

;; Calculate for moving pc and viewc forward (dx > 0) or back

(defun Vfb (dx)
 (declare (type single-float dx))
  (let ((delta-v (v*-constant vipc (- dx))))
    (setq pc (v+ pc delta-v))
    (setq viewc (v+ viewc delta-v))
  )
)

;; Characters specifying commands

(defconstant Vchars "pnlrfbPNLRFBudzZeEqQ")

;; Check for numeric parameter, returns (float) prm or NIL

(defun vnumchk (prm)
  (when (and prm (not (numberp prm)))
    (format t "Bad parameter -- ignored.~%") (setq prm NIL))
  (if prm (+ prm 0.0) prm)
)

;; Check for vector parameter

(defun vvectchk (prm)
  (when (and prm
            (not (and (vectorp prm) (eql (length prm) 3)
		      (numberp (aref prm 3)))))
    (format t "Bad parameter -- ignored.~%") (setq prm NIL))
  prm
)

;; Decode command, null parameter means retain current value.
;; Returns MOVE, T or QUIT.  Calculates new view but neither
;; adjusts projections nor plots new view.

(defun vperf (cm prm)
 (declare (type character cm))
  (viewstate)
  (cond ((find cm "pnlrPNLR")
	 (setq prm (vnumchk prm))
         (if prm (setq deltath (setq prm (/ prm 57.2957795)))
                 (setq prm deltath))
         (case cm
           (#\p (pupdn prm))  (#\n (pupdn (- prm)))
           (#\r (prtlf prm))  (#\l (prtlf (- prm)))
           (#\P (Vupdn prm))  (#\N (Vupdn (- prm)))
           (#\R (Vrtlf prm))  (#\L (Vrtlf (- prm))))
         'MOVE)
        ((find cm "fbFB")
	 (setq prm (vnumchk prm))
         (if prm (setq deltax prm) (setq prm deltax))
         (case cm
           (#\f (pfb prm))  (#\b (pfb (- prm)))
           (#\F (Vfb prm))  (#\B (Vfb (- prm))))
         'MOVE)
	((eql cm #\u)
	 (setq prm (vvectchk prm))
	 (when prm (setq viup (vscale-to-unit-vector prm)) 'MOVE))
	((find cm "sS")
         (setq viewc std-viewc)
	 (setq pc    std-pc)
	 (setq viup  std-viup)
	 'MOVE)
        ((find cm "wW") (report-view) T)
	((find cm "zZ")
	 (setq pc (v- pc viewc)) (setq viewc #(0.0 0.0 0.0)) 'MOVE)
        ((eql cm #\d)
         (setq prm (vnumchk prm))
         (when prm (setq lumdn2 (* prm prm)) 'MOVE))
        ((find cm "eE") (print (eval prm)) T)
        ((find cm "qQ") 'QUIT)
        (T (format T "Unrecognized command~%") T))
)

;; Report current status:

(defun report-view ()
 (let ((*print-array* T))
      (format T "viewc = ~A  pc = ~A  viup = ~A~%" viewc pc viup)
 )
)

;; Test for searching strings

(defun white (c) (eql c #\space))

;; Interpret string containing command

(defun vinter (cstr)
 (let* ((cpsn (position-if-not #'white cstr))
        (cmdc (char cstr cpsn))
        (cprm (if (>= (1+ cpsn) (length cstr)) NIL
                    (read-from-string cstr NIL NIL :start (1+ cpsn))))
        (cval (vperf cmdc cprm)))
   (if (eq cval 'MOVE) (setup-proj pc pn viup))
   cval
 ) ; end let*
)

;; Read and obey view commands until q (Q).    Commands are one line.

(defun view ()
 (prog (v)
    (viewstate)
    (setq std-viewc viewc) (setq std-pc pc) (setq std-viup viup)
   LOOP
    (format t "view> ")
    (setq v (vinter (read-line)))
    (if (eq v 'MOVE) (plot-galaxies))
    (if (not (eq v 'QUIT)) (go LOOP))
 )
)

;; Function to check for input and interpret as view command
;; if present.  Updates projection, doesn't plot.  Returns QUIT
;; if command indicates quitting, T or MOVE otherwise if command
;; found, NIL if no input ready.

(defun check-view ()
 (let ((c (read-char-no-hang)))
   (when c (unread-char c) (setq c (vinter (read-line))))
   c
 )
)
